%%% counter.pl
%% a counter is a pair (Major/Minor) of non-negative integers.
%% Major is bounded to be less or equal than the number of levels in
%% the semantic tree. 
%% Likewise, Minor is bounded to be less or equal than the number of
%% sublevels of the Major level it belongs to.
%% 
%% Operations:
%% clear: reset all major/minor counters to 0
%% set_major: set counter major value
%% set_minor: set counter minor value for given major number
%% next: return the least value Major/Minor in the lexicographical ordering 
%%       that is smaller than the global Level/Sublevel counter
%% max: return max value
%% inc: increment 
%%
%% Currently, the counter name is part of the procedures.
%% ToDo: a more systematic approach (macros?)
%%

:-  make_array(univ_extend_ctr(1000),integer).
    % The node with highest value of SubLevel for which all
    % unit conclusions have been computed


univ_extend_ctr_clear :-
	setval(next_univ_extend_ctr,0),
	repeat, 
	getval(next_univ_extend_ctr,SLevel),
	getval(level,Level),
	( SLevel > Level ->
	  true
        ; setval(univ_extend_ctr(SLevel),0),
	  incval(next_univ_extend_ctr),
	  fail
        ).


univ_extend_ctr_set_major(N) :-
	setval(next_univ_extend_ctr,N).

univ_extend_ctr_set_minor(Level,N) :-
	setval(univ_extend_ctr(Level),N).

univ_extend_ctr_get_minor(Level,N) :-
	getval(univ_extend_ctr(Level),N).

univ_extend_ctr_next(ResLevel+ResSubLevel) :-
	getval(next_univ_extend_ctr,CLevel),
	getval(level,Level), !, % green cut
	( CLevel =< Level ->
	  getval(univ_extend_ctr(CLevel),CLevelSatMax),
	  getval(sublevel(CLevel),CLevelTotal), !, % green cut
	  ( CLevelSatMax =< CLevelTotal ->
	    ResLevel = CLevel,
	    ResSubLevel = CLevelSatMax
	  ; incval(next_univ_extend_ctr),
%	    getval(next_univ_extend_ctr,NLevel),
%	    setval(univ_extend_ctr(NLevel),0), !, % green cut
	    univ_extend_ctr_next(ResLevel+ResSubLevel)
	  )
        ), !.

univ_extend_ctr_max(Level+SubLevel) :-
	getval(level,Level),
	getval(univ_extend_ctr(Level),SubLevel).

univ_extend_ctr_inc :-
	getval(next_univ_extend_ctr,Level),
	incval(univ_extend_ctr(Level)).


