
:- dynamic(ptemplate/1),
   dynamic(template/1),
   dynamic(inf_rule/1),
   dynamic(dp_flag/2),
   dynamic(ground_clause_set/0),
   dynamic(equality_clause_set/0),
   dynamic(function_free_clause_set/0),
   dynamic(constant/1),
   dynamic(soi_clause/1),
   dynamic(tmp_soi_clause/2),
   dynamic(horn_clause_set/0),
   dynamic(pred_sym/1).

init_parser(ReadFlags) :-
	setval(read_flags,ReadFlags),
	retract_all(soi_clause(_)),
	retract_all(tmp_soi_clause(_,_)),
	retract_preds_dynamic,
	retract_all(ptemplate(_)),
	( getval(read_flags, true) -> 
	  retract_all(dp_flag(_,_))
        ; true 
        ),
	retract_all(template(_)),
	retract_all(constant(_)),
	retract_all(ground_clause_set),
	retract_all(equality_clause_set),
	retract_all(horn_clause_set),
	retract_all(function_free_clause_set),
	setval(max_clause_length,0),
	assert(ground_clause_set),
	assert(function_free_clause_set),
        assert(horn_clause_set),
	retract_all(inf_rule(_)).


consult_file(File) :-

	(string(File) -> TmeFilename = File
    ;
	concat_atoms(File, '.tme', TmeFilename) ),

	open(TmeFilename, read, TmeFile),
	
%	write('% Reading '), write(TmeFilename), writeln('...'), nl,
	repeat,
	read(TmeFile, Term),
%	write('% '), writeln(Term),
	(translate(Term) -> true),
	Term == end_of_file,
	!,

	%assert(constant(any_term)),
	close(TmeFile),
	sort_clause_set,
%	write('% '), write(TmeFilename), writeln(' done.'), nl,
        true.


translate(end_of_file) :- !.

%translate(mission_flag(Name, Value) ) :- !, flag(Name, Value).

translate(read(File) ) :- !, consult_file(File).

translate(Term) :- 

	(protein_term(Term) ; mission_term(Term) ; ht_term(Term)),
	!, 
	nl.


translate(Term) :- 

	dp_term(Term), 
	tr_dp(Term),
	!.



translate( (?- Goal) ) :- !,
	pl_list(Goal, GoalList),
	negate_list(GoalList, SOI_clause), 
	normalize_assert(SOI_clause).

translate( (P -> C) ) :- !,
	pl_list(P, PL),
	norm_list(PL,PLN),
	norm_list([C],[CN]),
	(mydelete((-Goal),PLN,Rest)
          -> assert(inf_rule((-Goal, Rest -> CN)))
          ; true).

translate( (Head :- Body) ) :- !,

	pl_list(Head, HeadList), 
	pl_list(Body, BodyList), 

	negate_list(BodyList, NBodyList), 
	append(HeadList, NBodyList, SOI_clause),
	normalize_assert(SOI_clause).

translate(Clause) :-  % should be ';' separated list of literals

	pl_list(Clause, SOI_clause), 
	normalize_assert(SOI_clause).

ht_term(ht_flag(_,_)).
dp_term(dp_flag(_,_)).

%% used for flags:
tr_dp(dp_flag(Flag,Value)) :- 
	\+ getval(read_flags, true), 
	% so we skip reading of flags
	!.
tr_dp(dp_flag(Flag,Value)) :- 
	FPattern =.. [dp_flag,Flag,_],
	retract_all(FPattern),
	assert(dp_flag(Flag,Value)).

protein_term(protein_flag(_, _)).
protein_term(pttp_flag(_, _)).
protein_term(calculus(_)).
protein_term(mod_flag(_,_)).
protein_term(begin(_)).
protein_term(end(_)).

mission_term(mission_flag(_, _)).


analyze_clause([]).
analyze_clause([-Lit|Rest]) :- !,
	analyze_lit(Lit), 
	analyze_clause(Rest), !.
analyze_clause([Lit|Rest]) :-
	analyze_lit(Lit), 
	analyze_clause(Rest), !.

analyze_lit(L) :-
	functor(L,F,_N),
	% constraints contain literals, so don't make terms out of them
	% during analyzing
	(F == unequal ; F == (~=)).

analyze_lit(L) :-
	functor(L,F,N),
	(pred_sym(F / N) ; assert(pred_sym(F / N))),
	L =.. [_P|Args],
	analyze_term_list(Args).

analyze_term_list([]).
analyze_term_list([T|R]) :-
        analyze_term(T),
	analyze_term_list(R).

analyze_term(X) :- var(X),         
	retract_all(ground_clause_set),
	!.
analyze_term(T) :- 
	functor(T,F,N),
	(N = 0 ->
	    (constant(F) ; assert(constant(F)))
         ;  retract_all(function_free_clause_set),
	    template_args(N,TemplateArgs),
	    Template =.. [F|TemplateArgs],
	    (template(Template) ; assert(template(Template)))),
	T =.. [_F|Args],
	analyze_term_list(Args).
	     

template_args(0,[]).
template_args(1,[_]).
template_args(2,[_,_]).
template_args(3,[_,_,_]).
template_args(4,[_,_,_,_]).
template_args(5,[_,_,_,_,_]).
template_args(6,[_,_,_,_,_,_]).
template_args(7,[_,_,_,_,_,_,_]).
template_args(8,[_,_,_,_,_,_,_,_]).

normalize_assert(Clause) :-
        norm_clause(Clause,NC),
	analyze_clause(NC),
	add_pos_sign(NC,SNC),
	( \+ horn_clause_set ->
	  true
        ; ( delete(+_,SNC,SNC1),
	    member(+_,SNC1) ) ->
	  retract_all(horn_clause_set)
        ; true
        ),
	length(SNC,SNCLen),
	( ( \+ equality_clause_set,
	    % see if we have a clause for symmetry:
	    SNCLen = 2,
	    ( variant([+(X=Y), -(Y=X)],SNC) ;
	      variant([-(X=Y), +(Y=X)],SNC) ) ) ->
	  assert(equality_clause_set)
        ; true 
        ),
	assert(tmp_soi_clause(SNC,SNCLen)),
	getval(max_clause_length,Max),
	max(Max,SNCLen,NewMax),
	setval(max_clause_length,NewMax).

sort_clause_set :-
	getval(max_clause_length,Max),
%	sort_clause_set(Max,0).
	M is Max + 1,
	sort_clause_set(1,M).

sort_clause_set(Bound,Bound).
sort_clause_set(Current,Bound) :-
	repeat,
	  \+ ( retract(tmp_soi_clause(C,Current)),
	       assert(soi_clause(C))
	     ),
%	Next is Current - 1,
	Next is Current + 1,
	sort_clause_set(Next,Bound).
	

add_pos_sign([],[]).
add_pos_sign([-F|R],[-F|SR]) :-
	!, add_pos_sign(R,SR).
add_pos_sign([F|R],[+F|SR]) :-
	add_pos_sign(R,SR).

polarity_switch([],[]).
polarity_switch([L|R],[L1|R1]) :- negate(L,L1), polarity_switch(R,R1).

premisses( (Prem, More), Norm) :- !,

	positive_literal(Prem, Next, Norm),
	premisses(More, Next).

premisses(Prem, Norm) :- positive_literal(Prem, [], Norm).


conclusio( (Conc, More), Pres, Norm) :- !,

	negative_literal(Conc, Next, Norm),
	conclusio(More, Pres, Next).

conclusio( (Conc; More), Pres, Norm) :- !,

	negative_literal(Conc, Next, Norm),
	conclusio(More, Pres, Next).

conclusio(Conc, Pres, Norm) :- negative_literal(Conc, Pres, Norm).


positive_literal(true, Norm, Norm) :- !.

positive_literal(~ Lit, More, [- Lit | More]) :- !.

positive_literal(Lit, More, [Lit | More]).


negative_literal(false, Norm, Norm) :- !.

negative_literal(~ Lit, More, [Lit | More]) :- !.

negative_literal(Lit, More, [Neg | More]) :- negate(Lit, Neg).


retract_preds_dynamic :-
	pred_sym(P/N),
%	template_args(N,L),
%	Pred =.. [P|L],
%	retract_all(Pred), 
	retract(pred_sym(P/N)),
	fail.
retract_preds_dynamic.

make_preds_dynamic :-
	pred_sym(P/N),
	template_args(N,L),
	Pred =.. [P|L],
	translate((~Pred, Pred -> false)),
	fail.
make_preds_dynamic.


